# coding=utf8
""" Body Docs

Main entry into calling the rest documentation generator
"""

__author__		= "Chris Nasr"
__version__		= "1.0.0"
__copyright__	= "Ouroboros Coding Inc."
__email__		= "chris@ouroboroscoding.com"
__created__		= "2025-02-23"

# Python imports
from argparse import ArgumentParser
from sys import exit, stderr

# Constants
EXTENSIONS = { 'markdown': 'md' }

# Local impors
from . import handle_file, generate_service

def cli() -> int:
	"""CLI

	Handles generating a service's documentation via the command line interface

	Returns:
		int
	"""

	# Setup the argument parser
	oArgParser = ArgumentParser(
		description='Body - Generate REST Documentation'
	)
	oArgParser.add_argument(
		'-f', '--format', nargs=1, default='markdown',
		help='The format to out the documentation in. Only supports ' \
			'"markdown" at present'
	)
	oArgParser.add_argument(
		'-o', '--output', default='./',
		help='The folder to store generated documentation in'
	)
	oArgParser.add_argument(
		'-p', '--parser', nargs=1, default='google',
		help='The parser to use for docstrings. Only supports "google" at ' \
			'present'
	)
	oArgParser.add_argument(
		'file', nargs="+",
		help='The file(s) to parse and generate documentation from'
	)
	oArgs = oArgParser.parse_args()

	# If the format is anything other than markdown
	if oArgs.format != 'markdown':
		print('Invalid option for --format: "%s"' % oArgs.format,
			file = stderr)
		print('body-docs only supports "markdown" generator')
		return 0

	# If the parser is anything other than markdown
	if oArgs.parser != 'google':
		print('Invalid option for --parser: "%s"' % oArgs.format,
			file = stderr)
		print('body-docs only supports "google" docstring format')
		return 0

	# Init the list of services
	lServices = []

	# Go through each file passed
	for sFile in oArgs.file:

		# Handle the file
		mRes = handle_file(sFile, oArgs.parser)
		if not mRes:
			print('Failed to handle file "%s"' % sFile, file = stderr)
			return 1

		# Extend the list
		lServices.extend(mRes)

	# If we have services
	if lServices:

		# Generate a file for each
		for dService in lServices:
			print('Generating "%s"' % dService['name'])
			if not generate_service(
				dService,
				oArgs.format,
				oArgs.output.rstrip('/')
			):
				print('Failed to generate "%s"' % dService['name'],
		  			file = stderr)
				return 1

	# Return OK
	return 0

# Only run if called directly
if __name__ == '__main__':
	exit(cli())


# coding=utf8
""" Parser

Parses a method's docstring to get all the parts
"""

__author__		= "Chris Nasr"
__version__		= "1.0.0"
__copyright__	= "Ouroboros Coding Inc."
__email__		= "chris@ouroboroscoding.com"
__created__		= "2025-02-23"

# Python imports
import re

# Regular expressions
_google_section = re.compile(r'^([^\s:]+(?: [^\s:]+)*):\s*(.*)$')
_google_whitespace = re.compile(r'^(\s+)')

def parse_google(s: str) -> dict:
	"""Google

	Parsers google style docstring and returns the sections found

	Arguments:
		s (str): The docstring to parse

	Returns:
		dict
	"""

	# Split the docstring into lines
	lLines = s.split('\n')

	# Init the return with the name
	dSections = {
		'name': lLines[0].strip(),
		'description': []
	}

	# Start with the description
	sKey = 'description'

	# Init the whitespace regex
	oWhitespace = None

	# Loop through the remaining lines
	for sLine in lLines[1:]:

		# If the line is a new section header
		m = _google_section.match(sLine)
		if m:

			# If we need to skip it
			if sLine[0] == '-' and sLine[-1] == '-':
				sLine = sLine[1:-1]

			# Else, continue processing the new section
			else:

				# Check the last line of the previous section, if it's an empty line
				#	delete it
				if len(dSections[sKey]) and dSections[sKey][-1].strip() == '':
					dSections[sKey].pop()

				# If the previous section is empty
				if not dSections[sKey]:
					del dSections[sKey]

				# Clear the whitespace
				oWhitespace = None

				# Store the new section name
				sKey = m.group(1).lower().replace(' ', '_')
				dSections[sKey] = []

				# If we have a group 2, set it as the new line
				if m.group(2) is not None:
					sLine = m.group(2)

				# Else, loop back around to the next line
				else:
					continue

		# If we have nothing in the section yet
		if not dSections[sKey]:

			# If the line is empty, skip it
			if sLine.strip() == '':
				continue

			# If the line has whitespace
			m = _google_whitespace.match(sLine)
			if m:
				oWhitespace = re.compile('^(?:%s)?(.*)' % m.group(1))

		# If we have whitespace
		if oWhitespace:

			# Strip it off
			m = oWhitespace.match(sLine)
			sLine = m.group(1)

		# Add the line to the current section
		dSections[sKey].append(sLine)

	# Check the last section
	if not dSections[sKey]:
		del dSections[sKey]

	# Go through each section and turn lines into a string
	for s in dSections:
		if isinstance(dSections[s], list):
			dSections[s] = '\n'.join(dSections[s])

	# Return the sections
	return dSections


# coding=utf8
""" Regex

Shared regular expressions
"""

__author__		= "Chris Nasr"
__copyright__	= "Ouroboros Coding Inc."
__email__		= "chris@ouroboroscoding.com"
__created__		= "2023-01-25"

# Python imports
import re

EMAIL_ADDRESS = re.compile(r'^[^@\s]+@[^@\s]+\.[a-zA-Z0-9]{2,}$')
"""Email Address"""

PHONE_NUMBER_NA = re.compile(r'^\+?1?[ -]?\(?(\d{3})\)?[ -]?(\d{3})[ -]?(\d{4})$')
"""North American Phone Number"""


# coding=utf8
""" Docs

Generates markdown file of all request points in a service
"""

__author__		= "Chris Nasr"
__version__		= "1.0.0"
__copyright__	= "Ouroboros Coding Inc."
__email__		= "chris@ouroboroscoding.com"
__created__		= "2025-02-23"

# Ouroboros imports
from strings import to_file

# Python imports
import ast
from os import path
import re
from sys import exit, stderr
from typing import List, Literal

# Local imports
from .parser import parse_google

# Regular expressions
_service = re.compile(r'request:\s*(\w+)', re.IGNORECASE)
_data = re.compile(
	r'^([^,]+)\s*,\s*([^,]+)\s*,\s*(no|yes)\s*,\s*(.+)$', re.IGNORECASE
)
_error = re.compile(r'^(\d+)\s*,\s*((?:[a-z0-9_]+\.)?[A-Z0-9_]+)\s*,\s*(.+)$')
_response = re.compile(r'^([^,]+)\s*,\s*([^,]+)\s*,\s*(.+)$')

# Constants
EXTENSIONS = { 'markdown': 'md' }

def handle_class(
	_class: ast.ClassDef,
	parser: Literal['google']
) -> dict | Literal[False]:
	"""Handle Class

	Looks through a Service class and generates the data for the documentation

	Arguments:
		_class (ast.ClassDef): The AST definition of the class we will step
			through
		parser ('google'): The parser to use for the docstrings

	Returns:
		dict | False
	"""

	# If it's not a Service class
	if not any([ o.id == 'Service' for o in _class.bases ]):
		return False

	# Init the return
	dRet = {
		'name': _class.name,
		'file': _class.name.lower(),
		'uri': _class.name.lower(),
		'requests': [ ]
	}

	# Get the docstring and store it
	sDoc = ast.get_docstring(_class, True)

	# Parse the docstring to get the sections
	if parser == 'google':
		dSections = parse_google(sDoc)
	else:
		raise ValueError('parser', 'invalid parser "%s"' % parser)

	# If we have a name override
	if 'docs-file' in dSections:
		dRet['file'] = dSections['docs-file']
	elif 'docs_file' in dSections:
		dRet['file'] = dSections['docs_file']

	# Add the description
	dRet['description'] = dSections['description']

	# If we have a body overwrite
	if 'docs-body' in dSections:
		dRet['body'] = dSections['docs-body']
	elif 'docs_body' in dSections:
		dRet['body'] = dSections['docs_body']

	# Step through each child of the class
	for oMethod in _class.body:

		# If it's not a function
		if not isinstance(oMethod, ast.FunctionDef):
			continue

		# Handle the method
		mRes = handle_method(oMethod, parser)
		if mRes:
			dRet['requests'].append(mRes)

	# Return OK
	return dRet

def handle_file(
	file: str,
	parser: Literal['google']
) -> List[dict] | Literal[False]:
	"""Handle File

	Looks through a single file for Service classes. Returns False is anything
	fails, else a list of services found in the file

	Arguments:
		file (str): The path to the file
		parser ('google'): The parser to use for the docstrings

	Returns:
		dict[] | False
	"""

	# If the file does not exist
	if not path.exists(file):
		print('File "%s" does not exist' % file, file = stderr)
		return False

	# Open the file
	with open(file) as oF:

		# Get the abstract syntax tree for the file
		try:
			oAST = ast.parse(oF.read(), file)

			# Init list of services
			lServices = []

			# Go through each node in the tree
			for oClass in ast.iter_child_nodes(oAST):

				# If it's a class
				if isinstance(oClass, ast.ClassDef):

					# Handle the class
					mRes = handle_class(oClass, parser)
					if mRes:
						lServices.append(mRes)

			# If there's no services
			if not lServices:
				return False

			# Return OK
			return lServices

		# Catch syntax errors from broken code
		except SyntaxError as e:
			print('Syntax Error parsing "%s" at line %d, column %d\n' % (
				file,
				e.args[1][1],
				e.args[1][2]
			), file = stderr)
			return False

	# Strange
	print('Unknown issue reading "%s"' % file, file = stderr)
	return False

def handle_method(
	method: ast.FunctionDef,
	parser: Literal['google']
) -> dict | Literal[False]:
	"""Handle Method

	Looks through a service method and generates the data for the documentation

	Arguments:
	 	method (ast.FunctionDef): The AST definition of the method
		parser ('google'): The parser to use for the docstrings

	Returns:
		dict | False
	"""

	# Split the name and store the parts
	lParts = method.name.split('_')

	# If it's not a valid action
	if lParts[-1] not in [ 'create', 'delete', 'read', 'update' ]:
		return False

	# If it doesn't have the expected arguments
	try:
		lArgs = [ o.arg for o in method.args.args ]
		if lArgs != [ 'self', 'req' ]:
			return False
	except Exception:
		return False

	# Init the return
	dRet = {
		'action': lParts[-1],
		'noun': ' '.join(lParts[:-1]),
		'uri': '/'.join(lParts[:-1])
	}

	# Get the docstring and store it
	sDoc = ast.get_docstring(method, True)

	# Parse the docstring to get the sections
	if parser == 'google':
		dSections = parse_google(sDoc)
	else:
		raise ValueError('parser', 'invalid parser "%s"' % parser)

	# Add the name and description
	dRet['name'] = dSections['name']
	dRet['description'] = 'description' in dSections and \
		dSections['description'] or ''

	# If we have a data section
	if 'data' in dSections:

		# Split data into lines, strip whitespace
		lLines = [ s.strip() for s in dSections['data'].split('\n') ]

		# Init data
		lData = []

		# Go through each line
		for s in lLines:

			# Match it
			m = _data.match(s)
			if not m:
				print('Invalid "data" line in "%s" docstring: %s' % (
					method.name, s
				), file = stderr)
			else:
				lData.append({
					'name': m.group(1),
					'type': m.group(2),
					'optional': m.group(3),
					'descr': m.group(4)
				})

		# If we have any, add it
		if lData:
			dRet['data'] = lData

		# Else, just pass the original string
		else:
			dRet['data'] = dSections['data']

	# If we have a response section
	if 'response' in dSections:

		# Split response into lines, strip whitespace
		lLines = [ s.strip() for s in dSections['response'].split('\n') ]

		# Init response
		lResponse = []

		# Go through each line
		for s in lLines:

			# Match it
			m = _response.match(s)
			if m:
				lResponse.append({
					'name': m.group(1),
					'type': m.group(2),
					'descr': m.group(3)
				})

		# If we have any, add it
		if lResponse:
			dRet['response'] = lResponse

		# Else, just pass the original string
		else:
			dRet['response'] = dSections['response']

	# If we have a response example
	if 'response_example' in dSections:
		dRet['response_example'] = dSections['response_example']

	# If we have an error section
	if 'error' in dSections:

		# Split error into lines, strip whitespace
		lLines = [ s.strip() for s in dSections['error'].split('\n') ]

		# Init error
		lError = []

		# Go through each line
		for s in lLines:

			# Match it
			m = _error.match(s)
			if not m:
				print('Invalid "error" line in "%s" docstring: %s' % (
					method.name, s
				), file = stderr)
			else:
				lError.append({
					'code': m.group(1),
					'const': m.group(2),
					'descr': m.group(3)
				})

		# If we have any, add it
		if lError:
			dRet['error'] = lError

	# If we have an example section
	if 'example' in dSections:
		dRet['example_raw'] = dSections['example']
	elif 'data_example' in dSections:
		dRet['data_example'] = dSections['data_example']

	# Return the info
	return dRet

def generate_service(service: dict, format: str, output: str):
	"""Generate Service

	Generates a single file with all documentation for the service

	Arguments:
		service (dict): The service info with the nouns
		format ('markdown'): The format used to generate the docs
		output (str): The directory to store the docs in
	"""

	# Load and init Jinja2
	try:
		from jinja2 import Environment, PackageLoader, select_autoescape
		oJinja = Environment(
			autoescape = select_autoescape(),
			loader = PackageLoader('body.docs'),
			lstrip_blocks = True,
			trim_blocks = True
		)

	# Notify the user to install Jinja2 in development
	except ModuleNotFoundError as e:
		print('Using "body-docs" requires that Jinja2 be installed. It is ' \
			'not installed by default to save space on production installs.\n' \
			'To install jinja2 run the following in your venv:\n\n' \
			'pip install jinja2', file = stderr)
		exit(1)

	# Get the template based on the format
	oTpl = oJinja.get_template('%s.j2' % format)

	# Render the template and save it to the output folder
	return to_file(
		'%s/%s.%s' % ( output, service['file'], EXTENSIONS[format] ),
		oTpl.render(**service)
	)


# coding=utf8
"""Body

Shared methods for accessing the brain and other shared formats
"""

__author__		= "Chris Nasr"
__copyright__	= "Ouroboros Coding Inc."
__email__		= "chris@ouroboroscoding.com"
__created__		= "2022-08-29"

__all__ = [
	'create', 'read', 'update', 'delete',
	'Service', 'Error', 'Response', 'ResponseException'
]

# Import external calls and Service
from body.external import create, delete, read, update
from body.response import Error, Response, ResponseException
from body.service import Service